import inspect

from pandas_ml_utils import FeaturesAndLabels, Typing, Callable, List
from pandas_ml_utils.ml.data.extraction import extract_with_post_processor


class PostProcessedFeaturesAndLabels(FeaturesAndLabels):

    def __init__(self,
                 features: Typing._Selector,
                 feature_post_processor: List[Callable[[Typing.PatchedDataFrame], Typing.PatchedDataFrame]],
                 labels: Typing._Selector,
                 labels_post_processor: List[Callable[[Typing.PatchedDataFrame], Typing.PatchedDataFrame]] = None,
                 sample_weights: Typing._Selector = None,
                 sample_weights_post_processor: List[Callable[[Typing.PatchedDataFrame], Typing.PatchedDataFrame]] = None,
                 gross_loss: Typing._Selector = None,
                 gross_loss_post_processor: List[Callable[[Typing.PatchedDataFrame], Typing.PatchedDataFrame]] = None,
                 targets: Typing._Selector = None,
                 targets_post_processor: List[Callable[[Typing.PatchedDataFrame], Typing.PatchedDataFrame]] = None,
                 label_type=None,
                 **kwargs):
        super().__init__(
            PostProcessedFeaturesAndLabels.post_process(features, feature_post_processor),
            PostProcessedFeaturesAndLabels.post_process(labels, labels_post_processor),
            PostProcessedFeaturesAndLabels.post_process(sample_weights, sample_weights_post_processor),
            PostProcessedFeaturesAndLabels.post_process(gross_loss, gross_loss_post_processor),
            PostProcessedFeaturesAndLabels.post_process(targets, targets_post_processor),
            label_type,
            **kwargs
        )

        self._raw = [
            [features, feature_post_processor],
            [labels, labels_post_processor],
            [sample_weights, sample_weights_post_processor],
            [gross_loss, gross_loss_post_processor],
            [targets, targets_post_processor],
        ]

    @staticmethod
    def post_process(selectors, post_processors):
        # early exit if we actually do not post process
        if post_processors is None:
            return selectors

        # make post processors iterable and exceute one after the other
        pps = post_processors if isinstance(post_processors, list) else [post_processors]
        previous = selectors

        for pp in pps:
            previous = extract_with_post_processor(previous, pp)

        return previous

    def __repr__(self):
        def source(params):
            if params is None:
                return None
            else:
                return [inspect.getsource(p) if callable(p) else p if isinstance(p, (bool, int, float, str)) else repr(p) for p in params]

        return f'PostProcessedFeaturesAndLabels(' \
               f'\t{source(self._raw[0][0])}, ' \
               f'\t{source(self._raw[0][1])}, ' \
               f'\t{source(self._raw[1][0])}, ' \
               f'\t{source(self._raw[1][1])}, ' \
               f'\t{source(self._raw[2][0])}, ' \
               f'\t{source(self._raw[2][1])}, ' \
               f'\t{source(self._raw[3][0])}, ' \
               f'\t{source(self._raw[3][1])}, ' \
               f'\t{source(self._raw[4][0])}, ' \
               f'\t{source(self._raw[4][1])}, ' \
               f'\t{self.label_type}, ' \
               f'\t{self.kwargs}' \
               f')'



import pandas_ml_quant.analysis.normalizer as _norm
from pandas_ml_quant import Typing, has_indexed_columns
from arch import arch_model  # alternatively https://pyflux.readthedocs.io/en/latest/garch.html
import pandas as pd


def ta_garch11(df: Typing.PatchedPandas, period=200, forecast=1, returns='returns'):
    if has_indexed_columns(df):
        return pd.DataFrame(
            {f"{col}_garch11({period})->{forecast}": ta_garch11(df[col], period, forecast, returns) for col in df.columns},
            index=df.index
        )

    r = getattr(_norm, f'ta_{returns}')(df) if returns is not None else df

    def model(x):
        model = arch_model(x, p=1, q=1, dist='StudentsT', rescale=True)
        res = model.fit(update_freq=0, disp='off', show_warning=False)

        return res.forecast(horizon=forecast, method='analytic').variance.iloc[-1, -1] / (res.scale * res.scale)

    return r\
        .rolling(period)\
        .apply(model)\
        .rename(f"{df.name}_garch11({period})->{forecast}")




import tensorflow as tf
from keras import backend as _K
from keras.losses import categorical_crossentropy

from pandas_ml_common import LazyInit


def tailed_categorical_crossentropy(nr_of_categories, alpha=0.1, beta=1e10, delta=1, dtype='float32'):
    """
    assuming that we have discretized something like returns where we have less observations in the tails.
    If we want to train a neural net to place returns into the expected bucket we want to penalize if the
    prediction is too close to the mean. we rather want to be pessimistic and force the predictor to
    encounter the tails.

    :param nr_of_categories: number of categories aka length of the one hot encoded vectors
    :param alpha: describes the steepness of the parabola
    :param beta: used for the differentiable_argmax
    :param delta: is used to unevenly skew the loss to the outer bounds. 0 now skew > bigger skew
    :return: returns a keras loss function
    """

    parabolic_penalty = parabolic_crossentropy(nr_of_categories, delta, beta, dtype)

    # custom loss function is cross entropy with penalized tail errors
    def loss_function(y_true, y_pred):
        penalty = alpha * parabolic_penalty(y_pred, y_true)
        loss = categorical_crossentropy(y_true, y_pred)
        return loss + penalty

    return loss_function


def parabolic_crossentropy(nr_of_categories, delta=1, beta=1e10, dtype='float32'):
    argmax = DifferentiableArgmax(nr_of_categories, beta, dtype=dtype)
    offset = LazyInit(lambda: _K.constant(delta / 2, dtype=dtype))
    f = LazyInit(lambda: _K.constant((nr_of_categories + delta) / nr_of_categories, dtype=dtype))

    def loss_function(y_true, y_pred):
        return ((argmax(y_pred) + offset()) - (argmax(y_true)) * f()) ** 2

    return loss_function


def normal_penalized_crossentropy(nr_of_categories, alpha=10, beta=1e10, max_z=4, dtype='float32'):
    max_cat_int = nr_of_categories - 1
    _alpha = LazyInit(lambda: _K.constant(alpha, dtype=dtype))
    _mue = LazyInit(lambda: _K.constant(max_cat_int / 2, dtype=dtype))
    _sigma = LazyInit(lambda: _K.constant((max_cat_int - max_cat_int / 2) / abs(max_z), dtype=dtype))
    _argmax = DifferentiableArgmax(nr_of_categories, beta, dtype=dtype)
    _norm_dist = NormDistWeightedPenalizer(dtype)

    def loss_function(y_true, y_pred):
        # 1st we calculate the cross entropy
        loss = categorical_crossentropy(y_true, y_pred)

        # then we use the normal shaped probability distrubution to penalize the prediction
        penalty = _norm_dist(_argmax(y_pred), _mue(), _sigma()) * _alpha()

        return loss * penalty

    return loss_function


class DifferentiableArgmax(object):

    def __init__(self, nr_of_categories, beta=1e10, dtype='float32'):
        self.y_range = LazyInit(lambda: _K.arange(0, nr_of_categories, dtype=dtype))
        self.beta = beta

    def __call__(self, *args, **kwargs):
        y = args[0]
        return _K.sum(_K.softmax(y * self.beta) * self.y_range(), axis=-1)


class NormDistWeightedPenalizer(object):

    def __init__(self, dtype='float32'):
        self.one = LazyInit(lambda: _K.constant(1, dtype=dtype))
        self.two = LazyInit(lambda: _K.constant(2, dtype=dtype))
        self.pi = LazyInit(lambda: _K.constant(3.1415, dtype=dtype))
        self.e = LazyInit(lambda: _K.constant(2.7182, dtype=dtype))

    def __call__(self, *args, **kwargs):
        x, mu, sigma = args
        one, two, pi, e = self.one(), self.two(), self.pi(), self.e()

        # 1 / (sqrt(2pi) * sigma) * e ** -((x - mu)**2 / 2sigma**2)
        return  (one / (_K.sqrt(two * pi) * sigma)) * e ** -((x - mu) ** two / two * sigma ** two)


def mdn_cost(y_t, y_p):
    mu = y_p[:, 0]
    sigma = y_p[:, 1]

    dist = tf.distributions.Normal(loc=mu, scale=sigma)
    return tf.reduce_mean(-dist.log_prob(y_t))



import pandas as _pd

# create convenient type hint
from pandas_ml_common import Typing as _t


def ta_decimal_year(df: _t.PatchedPandas):
    return (df.index.strftime("%j").astype(float) - 1) / 366 + df.index.strftime("%Y").astype(float)


def ta_week_day(po: _t.PatchedPandas):
    if not isinstance(po.index, _pd.DatetimeIndex):
        df = po.copy()
        df.index = _pd.to_datetime(df.index)
    else:
        df = po

    return (df.index.dayofweek / 6.0).to_series(index=po.index, name="dow")


def ta_week(po: _t.PatchedPandas):
    if not isinstance(po.index, _pd.DatetimeIndex):
        df = po.copy()
        df.index = _pd.to_datetime(df.index)
    else:
        df = po

    return (df.index.week / 52.0).to_series(index=po.index, name="week")



from unittest import TestCase

from pandas_ml_quant.analysis import ta_garch11
from pandas_ml_quant_test.config import DF_TEST
import numpy as np


class TestVolatilityForecast(TestCase):

    def test_garch11(self):
        df = DF_TEST[-220:].copy()

        garch11 = ta_garch11(df[["Close"]])

        print(garch11)
        np.testing.assert_almost_equal(garch11.iloc[-1, -1], 0.000040, 5)




from unittest import TestCase

import numpy as np

from pandas_ml_quant import pd
from pandas_ml_quant.analysis.encoders import ta_rnn
from pandas_ml_quant_test.config import DF_TEST


class TestRegressionLagging(TestCase):

    def test_auto_regression(self):
        df = DF_TEST.copy()
        rnn = ta_rnn(df[["Close", "Volume"]], [1, 2])

        self.assertEqual(6761, len(rnn))


    def test_auto_regression_simple(self):
        df = DF_TEST.copy()
        rnn, min_needed_data = ta_rnn(df[["Close", "Volume"]], [1, 2],
                                      return_min_required_samples=True)

        self.assertEqual(2, min_needed_data)
        np.testing.assert_array_almost_equal(
            np.array([[df["Close"].iloc[-2], df["Close"].iloc[-3],
                       df["Volume"].iloc[-2], df["Volume"].iloc[-3]]]),
            rnn[-1:].values, 0.0001)

    def test_lag_smoothing_nan(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        #                               1, 2, 3, 4, 5, 6, 7, 8, 9, 10]  # original
        #                                  1, 2, 3, 4, 5, 6, 7, 8, 9]   # lag 1
        #                                        1, 2, 3, 4, 5, 6, 7]   # lag 1 + shift 2
        #                                        ^                      # this is where the df starts

        """when lag smoothing is enabled using shift (which is introducing nan into the data frame)"""
        rnn, min_needed_data = ta_rnn(df[["featureA"]],
                                      feature_lags=[0, 1],
                                      lag_smoothing={1: lambda df: df["featureA"].shift(2)},
                                      return_min_required_samples=True)


        len_features = 10 - 1 - 2

        """then"""
        self.assertEqual(len(rnn), len_features)
        self.assertAlmostEqual(rnn[0, "featureA"].iloc[0], 4)
        self.assertAlmostEqual(rnn[1, "featureA"].iloc[0], 1.0)
        self.assertAlmostEqual(rnn[0, "featureA"].iloc[-1], 10)
        self.assertAlmostEqual(rnn[1, "featureA"].iloc[-1], 7.0)



from .single_object import *
from .multi_object import *

"""
this module basically re-implements all oscillators from TA-Lib:
  https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
"""

"""
TODO add this missing indicators

AROONOSC - Aroon Oscillator
real = AROONOSC(high, low, timeperiod=14)
Learn more about the Aroon Oscillator at tadoc.org.

CMO - Chande Momentum Oscillator
real = CMO(close, timeperiod=14)
Learn more about the Chande Momentum Oscillator at tadoc.org.

MFI - Money Flow Index
real = MFI(high, low, close, volume, timeperiod=14)
Learn more about the Money Flow Index at tadoc.org.

STOCH - Stochastic
slowk, slowd = STOCH(high, low, close, fastk_period=5, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=0)
Learn more about the Stochastic at tadoc.org.

"""


from unittest import TestCase

from pandas_ml_quant.analysis import ta_ma_ratio
from pandas_ml_quant_test.config import DF_TEST


class TestNormalizer(TestCase):

    def test_ma_ratio(self):
        df = DF_TEST

        ma_ration_scaled = ta_ma_ratio(df)
        print(ma_ration_scaled.columns)
        self.assertTrue(True)


import pandas as pd

import pandas_ml_quant.analysis as analysis
import pandas_ml_quant.trading.strategy.optimized as optimized_strategies
from pandas_ml_quant.df.plot import TaPlot


class TechnicalAnalysis(object):

    def __init__(self, df: pd.DataFrame):
        self.df = df

    def plot(self, rows=2, cols=1, figsize=(18, 10), main_height_ratio=4):
        return TaPlot(self.df, figsize, rows, cols, main_height_ratio)


# add wrapper to call all indicators on data frames
def wrapper(func):
    def wrapped(quant, *args, **kwargs):
        if isinstance(quant.df.index, pd.MultiIndex):
            # we need to call the function for each top level item and join the result back
            raise ValueError("not implemented")

        return func(quant.df, *args, **kwargs)

    return wrapped


# add indicators
for indicator_functions in [analysis, optimized_strategies]:
    for indicator_function in dir(indicator_functions):
        if indicator_function.startswith("ta_"):
            setattr(TechnicalAnalysis, indicator_function[3:], wrapper(getattr(indicator_functions, indicator_function)))






from unittest import TestCase

import torch as t
import torch.nn as nn
from torch.optim import Adam

from pandas_ml_quant.pytorch.custom_loss import SoftDTW
from pandas_ml_quant_test.config import DF_TEST
from pandas_ml_quant import PostProcessedFeaturesAndLabels
from pandas_ml_utils import AutoEncoderModel, FeaturesAndLabels
from pandas_ml_utils.ml.model.pytoch_model import PytorchModel


class TestCustomLoss(TestCase):

    def test_soft_dtw_loss(self):
        df = DF_TEST[["Close"]][-21:].copy()

        class LstmAutoEncoder(nn.Module):
            def __init__(self):
                super().__init__()
                self.input_size = 1
                self.seq_size=10
                self.hidden_size = 2
                self.num_layers = 1
                self.num_directions = 1

                self._encoder =\
                    nn.RNN(input_size=self.input_size, hidden_size=self.hidden_size, num_layers=self.num_layers,
                           batch_first=True)

                self._decoder =\
                    nn.RNN(input_size=self.hidden_size, hidden_size=self.input_size, num_layers=self.num_layers,
                           batch_first=True)

            def forward(self, x):
                # make sure to treat single elements as batches
                x = x.view(-1, self.seq_size, self.input_size)
                batch_size = len(x)

                hidden_encoder = nn.Parameter(t.zeros(self.num_layers * self.num_directions, batch_size, self.hidden_size))
                hidden_decoder = nn.Parameter(t.zeros(self.num_layers * self.num_directions, batch_size, self.input_size))

                x, _ = self._encoder(x, hidden_encoder)
                x = t.repeat_interleave(x[:,-2:-1], x.shape[1], dim=1)
                x, hidden = self._decoder(x, hidden_decoder)
                return x

            def encoder(self, x):
                x = x.reshape(-1, self.seq_size, self.input_size)
                batch_size = len(x)

                with t.no_grad():
                    hidden = nn.Parameter(
                        t.zeros(self.num_layers * self.num_directions, batch_size, self.hidden_size))

                    # return last element of sequence
                    return self._encoder(t.from_numpy(x).float(), hidden)[0].numpy()[:,-1]

            def decoder(self, x):
                x = x.reshape(-1, self.seq_size, self.hidden_size)
                batch_size = len(x)

                with t.no_grad():
                    hidden = nn.Parameter(
                        t.zeros(self.num_layers * self.num_directions, batch_size, self.input_size))
                    return self._decoder(t.from_numpy(x).float(), hidden)[0].numpy()

        model = AutoEncoderModel(
            PytorchModel(
                PostProcessedFeaturesAndLabels(df.columns.to_list(), [lambda df: df.ta.rnn(10)],
                                               df.columns.to_list(), [lambda df: df.ta.rnn(10)]),
                LstmAutoEncoder,
                SoftDTW,
                Adam
            ),
            ["condensed-a", "condensed-b"],
            lambda m: m.module.encoder,
            lambda m: m.module.decoder,
        )

        fit = df.model.fit(model, epochs=100)
        print(fit.test_summary.df)

        encoded = df.model.predict(fit.model.as_encoder())
        print(encoded)



from typing import Tuple, Union

import torch as t
import torch.nn as nn
from torch.nn import init


class Time2Vec(nn.Module):
    """
    source: https://towardsdatascience.com/time2vec-for-time-series-features-encoding-a03a4f3f937e
    and:    https://arxiv.org/pdf/1907.05321.pdf
    """

    def __init__(self, in_dim: Union[Tuple[int, int], int], out_dim: int):
        super().__init__()
        if not isinstance(in_dim, tuple):
            in_dim = (in_dim, 1)

        self.output_dim = out_dim
        self.w = nn.Parameter(t.Tensor(1, 1))
        self.b = nn.Parameter(t.Tensor(in_dim[0], 1))
        self.W = nn.Parameter(t.Tensor(out_dim * in_dim[1], out_dim * in_dim[1]))
        self.B = nn.Parameter(t.Tensor(in_dim[0], out_dim * in_dim[1]))

    def reset_parameters(self):
        init.uniform_(self.W, 1e-6, 1)
        init.uniform_(self.B, 1e-6, 1)
        init.uniform_(self.w, 1e-6, 1)
        init.uniform_(self.b, 1e-6, 1)

    def forward(self, x):
        if x.ndim < 3:
            x = x.view(*x.shape, 1)

        original = self.w * x + self.b
        x = t.repeat_interleave(x, self.output_dim, dim=-1)
        sin_trans = t.sin((x @ self.W) + self.B)
        return t.cat([sin_trans, original], -1)




import pandas as _pd
from typing import Union as _Union

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_future_pct_to_current_mean(df: _pd.Series, forecast_period=1, period=14):
    future = df.shift(-forecast_period)
    mean = df.rolling(period).mean()

    return (future / mean) - 1




from unittest import TestCase

from pandas_ml_quant import PostProcessedFeaturesAndLabels, Constant
from pandas_ml_quant_test.config import DF_TEST


class TestFeaturePostProcesor(TestCase):

    def test_repr(self):
        fnl = PostProcessedFeaturesAndLabels(
            features=[
                "Close",
                lambda df: df["Close"].ta.trix(),
            ],
            feature_post_processor=[
                lambda df: df.ta.rnn(5)
            ],
            labels=[
                Constant(0)
            ],
            targets=[
                lambda df: df["Close"]
            ],
        )

        print(repr(repr(fnl).replace('"', '\"')))

        self.assertEqual(
            'PostProcessedFeaturesAndLabels(\t[\'Close\', \'                lambda df: df["Close"].ta.trix(),\\n\'], \t[\'                lambda df: df.ta.rnn(5)\\n\'], \t[\'Constant(0)\'], \tNone, \tNone, \tNone, \tNone, \tNone, \t[\'                lambda df: df["Close"]\\n\'], \tNone, \tNone, \t{})',
            repr(fnl)
        )

    def test_feature_post_processing(self):
        df = DF_TEST.copy()

        (f, min), l, t, w, gl = df._.extract(
            PostProcessedFeaturesAndLabels(
                features=[
                    "Close",
                    lambda df: df["Close"].ta.trix(),
                ],
                feature_post_processor=[
                    lambda df: df.ta.rnn(5)
                ],
                labels=[
                    Constant(0)
                ],
                targets=[
                    lambda df: df["Close"]
                ],
            )
        )

        self.assertEqual((6674, 10), f.shape)
        self.assertEqual((6674, 1), l.shape)
        self.assertEqual((6674, 5, 2), f._.values.shape)

    def test_feature_and_label_post_processing(self):
        df = DF_TEST.copy()

        (f, min), l, t, w, gl = df._.extract(
            PostProcessedFeaturesAndLabels(
                features=[
                    "Close",
                    lambda df: df["Close"].ta.trix(),
                ],
                feature_post_processor=[
                    lambda df: df.ta.rnn(5),
                    lambda df: df.ta.rnn(3),
                ],
                labels=[
                    Constant(0)
                ],
                labels_post_processor=[
                    lambda df: df.ta.rnn(4),
                ],
                targets=[
                    lambda df: df["Close"]
                ],
            )
        )

        self.assertEqual((6672, 30), f.shape)
        self.assertEqual((6672, 4), l.shape)
        # FIXME self.assertEqual((6674, 3, 5, 2), f._.values.shape)

    def test_post_z_standardisation(self):
        # FIXME implement me ...
        pass


from unittest import TestCase
from pandas_ml_quant import pd
from pandas_ml_quant.analysis.labels import ta_opening_gap, ta_future_bband_quantile, ta_future_multi_bband_quantile,\
    ta_opening_gap_closed
from pandas_ml_quant_test.config import DF_TEST


class TestDescreteLabels(TestCase):

    def test_ta_opening_gap(self):
        df = pd.DataFrame({"Open": [0, 2, 2, 2],
                           "Close":   [1, 2, 3, 0]})

        label = ta_opening_gap(df)

        self.assertListEqual([2., 0., 1.], label.dropna().tolist())

    def test_ta_future_bband_quantile(self):
        df = DF_TEST.copy()

        label1 = ta_future_bband_quantile(df["Close"], include_mean=False).dropna()
        label2 = ta_future_bband_quantile(df["Close"], include_mean=True).dropna()

        self.assertEqual(2.0, label1.values.max())
        self.assertEqual(3.0, label2.values.max())

    def test_ta_future_multi_bband_quantile(self):
        df = DF_TEST.copy()

        label = ta_future_multi_bband_quantile(df["Close"], include_mean=False).dropna()

        self.assertEqual(0.0, label.values.min())
        self.assertEqual(8.0, label.values.max())

    def test_ta_opening_gap_closed(self):
        df = DF_TEST[-30:].copy()

        df["label"] = ta_opening_gap_closed(df, no_gap=-1)

        self.assertListEqual(
            df["label"].values.tolist(),
            [-1, -1, False, -1, -1, -1, False, False, -1, -1, False, -1, -1, True, -1, -1, False, -1, True, -1, -1, True, False, -1, False, -1, -1, -1, False, True]
        )




from typing import Union as _Union

# create convenient type hint
import numpy as _np
import pandas as _pd

from pandas_ml_common import Typing as _t
from pandas_ml_common.utils import has_indexed_columns
from pandas_ml_quant.utils import wilders_smoothing as _ws, with_column_suffix as _wcs
from pandas_ml_utils import Model
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_sma(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"sma_{period}", df.rolling(period).mean())


def ta_ema(df: _PANDAS, period=12) -> _PANDAS:
    return _wcs(f"ema_{period}", df.ewm(span=period, adjust=False, min_periods=period-1).mean())


def ta_wilders(df: _PANDAS, period=12) -> _PANDAS:
    if has_indexed_columns(df):
        resdf = _pd.DataFrame({}, index=df.index)
        for col in df.columns:
            s = df[col].dropna()
            res = _np.zeros(s.shape)
            _ws(s.values, period, res)
            resdf = resdf.join(_pd.DataFrame({col: res}, index=s.index))

        res = resdf
    else:
        res = ta_wilders(df.to_frame(), period).iloc[:, 0]

    return _wcs(f"wilders_{period}", res)


def ta_multi_bbands(s: _pd.Series, period=5, stddevs=[0.5, 1.0, 1.5, 2.0], ddof=1, include_mean=True) -> _PANDAS:
    assert not has_indexed_columns(s)
    mean = s.rolling(period).mean().rename("mean")
    std = s.rolling(period).std(ddof=ddof)
    df = _pd.DataFrame({}, index=mean.index)

    for stddev in reversed(stddevs):
        df[f'lower-{stddev}'] = mean - (std * stddev)

    if include_mean:
        df["mean"] = mean

    for stddev in stddevs:
        df[f'upper-{stddev}'] = mean + (std * stddev)

    return df


def ta_bbands(df: _PANDAS, period=5, stddev=2.0, ddof=1, include_mean=True) -> _PANDAS:
    mean = df.rolling(period).mean()
    std = df.rolling(period).std(ddof=ddof)
    most_recent = df.rolling(period).apply(lambda x: x[-1], raw=True)

    upper = mean + (std * stddev)
    lower = mean - (std * stddev)
    z_score = (most_recent - mean) / std
    quantile = _pd.Series(None, index=df.index)
    quantile[most_recent <= lower] = 0
    quantile[most_recent > lower] = 1
    quantile[most_recent > mean] = 2
    quantile[most_recent > upper] = 3

    if isinstance(mean, _pd.Series):
        upper.name = "upper"
        mean.name = "mean"
        lower.name = "lower"
        z_score.name = "z"
        quantile.name = "quantile"
    else:
        upper.columns = _pd.MultiIndex.from_product([upper.columns, ["upper"]])
        mean.columns = _pd.MultiIndex.from_product([mean.columns, ["mean"]])
        lower.columns = _pd.MultiIndex.from_product([lower.columns, ["lower"]])
        z_score.columns = _pd.MultiIndex.from_product([z_score.columns, ["z"]])
        quantile.columns = _pd.MultiIndex.from_product([z_score.columns, ["quantile"]])

    if include_mean:
        return _pd.DataFrame(upper) \
            .join(mean) \
            .join(lower) \
            .join(z_score) \
            .join(quantile) \
            .sort_index(axis=1)
    else:
        return _pd.DataFrame(upper) \
            .join(lower) \
            .join(z_score) \
            .join(quantile) \
            .sort_index(axis=1)


def ta_model(df: _t.PatchedDataFrame, model: _Union[Model, str], post_processors=[]) -> _t.PatchedDataFrame:
    if isinstance(model, str):
        model = Model.load(model)

    p = df.model.predict(model)[PREDICTION_COLUMN_NAME]

    # apply post processors
    for pc in post_processors:
        p = pc(p)

    return p



from typing import Union as _Union

# create convenient type hint
import numpy as _np
import pandas as _pd
from scipy.stats import zscore

from pandas_ml_common import Typing
from pandas_ml_common.utils import has_indexed_columns
from pandas_ml_quant.analysis.filters import ta_ema as _ema, ta_wilders as _wilders, ta_sma as _sma
from pandas_ml_quant.utils import wilders_smoothing as _ws, with_column_suffix as _wcs

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_macd(df: Typing.PatchedPandas, fast_period=12, slow_period=26, signal_period=9, relative=True) -> _PANDAS:
    if has_indexed_columns(df):
        res = _pd.DataFrame({}, index=df.index)
        for col in df.columns.to_list():
            d = ta_macd(df[col], fast_period, slow_period, slow_period, relative)
            d.columns = _pd.MultiIndex.from_product([[col], d.columns])
            res = res.join(d)

        res.columns = _pd.MultiIndex.from_tuples(res.columns.to_list())
        return res
    else:
        fast = _ema(df, fast_period)
        slow = _ema(df, slow_period)

        macd = (fast / slow - 1) if relative else (fast - slow)
        signal = _ema(macd, signal_period)
        hist = macd - signal
        suffix = f'{fast_period},{slow_period},{signal_period}'

        for label, frame in {f"macd_{suffix}": macd, f"signal_{suffix}": signal, f"histogram_{suffix}": hist}.items():
            frame.name = label

        macd = macd.to_frame() if isinstance(macd, _pd.Series) else macd
        return macd.join(signal).join(hist)


def ta_mom(df: _PANDAS, period=10, relative=True) -> _PANDAS:
    return _wcs(f"mom_{period}", df.pct_change(period) if relative else df.diff(period))


def ta_roc(df: _PANDAS, period=10) -> _PANDAS:
    return _wcs(f"roc_{period}", df.pct_change(period))


def ta_stddev(df: _PANDAS, period=5, nbdev=1, ddof=1, downscale=True) -> _PANDAS:
    return _wcs(f"stddev_{period}", (df.rolling(period).std(ddof=ddof) * nbdev) / (100 if downscale else 1))


def ta_rsi(df: _PANDAS, period=14):
    returns = df.diff()

    pos = _wilders(returns.clip(lower=0), period)
    neg = _wilders(_np.abs(returns.clip(upper=0)), period)

    return _wcs(f"rsi_{period}", pos / (pos + neg), returns)


def ta_apo(df: _PANDAS, fast_period=12, slow_period=26, exponential=False, relative=True) -> _PANDAS:
    fast = _ema(df, fast_period) if exponential else _sma(df, fast_period)
    slow = _ema(df, slow_period) if exponential else _sma(df, slow_period)
    apo = (fast / slow) if relative else (fast - slow)
    return _wcs(f"apo_{fast_period},{slow_period},{int(exponential)}", apo, df)


def ta_trix(df: _PANDAS, period=30) -> _PANDAS:
    return _wcs(f"trix_{period}", _ema(_ema(_ema(df, period), period), period).pct_change() * 100, df)


def ta_ppo(df: _pd.DataFrame, fast_period=12, slow_period=26, exponential=True) -> _PANDAS:
    fast = _ema(df, period=fast_period) if exponential else _sma(df, period=fast_period)
    slow = _ema(df, period=slow_period) if exponential else _sma(df, period=slow_period)
    return _wcs(f"ppo_{fast_period},{slow_period},{int(exponential)}", (fast - slow) / slow, df)


def ta_zscore(df: _PANDAS, period=20, ddof=1, downscale=True):
    res = df.rolling(period).apply(lambda c: zscore(c, ddof=ddof)[-1] / (4 if downscale else 1))

    return _wcs(f"z_{period}", res)


def ta_up_down_volatility_ratio(df: _PANDAS, period=60, normalize=True, setof_date=False):
    if isinstance(df, _pd.DataFrame):
        return df.apply(lambda col: ta_up_down_volatility_ratio(col, period, normalize))

    returns = df.pct_change() if normalize else df
    std = _pd.DataFrame({
        "std+": returns[returns > 0].rolling(period).std(),
        "std-": returns[returns < 0].rolling(period).std()
    }, index=returns.index).fillna(method="ffill")

    ratio = (std["std+"] / std["std-"] - 1).rename("std +/-")

    # eventually we can off set the date such that we can fake one continuous data frame
    if setof_date:
        # +7 -1 binds us approximately to the same week day
        ratio.index = ratio.index - _pd.DateOffset(days=(ratio.index[-1] - ratio.index[0]).days + 7 - 1)

    return ratio





from unittest import TestCase

from pandas_ml_quant_test.config import DF_TEST, DF_TEST_MULTI
import pandas_ml_quant
import numpy as np

print(pandas_ml_quant.__version__)


class TestBackTest(TestCase):

    def test_crossing_sma_strategy(self):
        df = DF_TEST["Close"].copy()

        bt = df.ta.sma(20)\
               .ta.cross(df.ta.sma(60))\
               .ta.backtest(df, lambda sig: (sig, 1 if sig > 0 else -1))

        print(bt)
        np.testing.assert_almost_equal(151.908073, bt["net"].iloc[-1])

    def test_crossing_pairs_strategy(self):
        df = DF_TEST_MULTI._["Close"].copy()
        correlation = df["Close", "spy"].rolling(60).corr(df["Close", "gld"])
        signal = correlation\
            .to_frame()\
            .apply(lambda v: [-1, 1] if v[0] < -0.70 else ([1, -1] if v[0] > 0.70 else [0, 0]),
                   result_type='expand',
                   axis=1)

        porfolios = signal.ta.backtest(df, lambda sig: (sig, 10 * sig))
        print(porfolios)

    def test_markowitz(self):
        df = DF_TEST_MULTI._["Close"].copy()
        weights = df.ta.markowitz(return_period=20)

        prices = df.loc[weights.index]
        weights.columns = prices.columns
        shares_for_one_dollar = weights / prices

        portfolios = shares_for_one_dollar.ta.backtest(df, lambda shares: shares * 100, lambda value: -0.01)

        print(portfolios)
        np.testing.assert_almost_equal(225.853351, portfolios["Close,spy", "net"].iloc[-1])
        np.testing.assert_almost_equal(284.032017, portfolios["Close,gld", "net"].iloc[-1])



from typing import Union as _Union

import numpy as _np
import pandas as _pd

import pandas_ml_quant.analysis.filters as _f
import pandas_ml_quant.analysis.indicators as _i
from pandas_ml_common import get_pandas_object as _get_pandas_object
from pandas_ml_quant.utils import index_of_bucket as _index_of_bucket

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_cross_over(df: _pd.DataFrame, a=None, b=None, period=1) -> _PANDAS:
    # return only > 0
    return ta_cross(df, a, b, period).clip(lower=0)


def ta_cross_under(df: _pd.DataFrame, a=None, b=None, period=1) -> _PANDAS:
    # return only < 0
    return ta_cross(df, a, b, period).clip(upper=0)


def ta_cross(df: _pd.DataFrame, a=None, b=None, period=1):
    # get pandas objects for crossing
    if a is None and b is None:
        assert len(df.columns) == 2, f"ambiguous crossing of {df.columns}"
        a = df[df.columns[0]]
        b = df[df.columns[1]]
    elif b is None:
        b = _get_pandas_object(df, a)
        a = df
    elif a is None:
        b = _get_pandas_object(df, b)
        a = df
    else:
        a = _get_pandas_object(df, a)
        b = _get_pandas_object(df, b)

    # get periods
    a1 = a.shift(period)
    b1 = b.shift(period)

    # if a1 < b1 and a > b then a crosses over b
    a_over_b = ((a1 < b1) & (a > b)).astype(int)

    # if a1 > b1 and a < b then a cross under b
    a_under_b = ((a1 > b1) & (a < b)).astype(int) * -1

    return a_over_b + a_under_b


def ta_future_crossings(df: _PANDAS, a=None, b=None, period=1, forecast=1):
    crossings = _i.ta_cross(df, a, b, period=period)

    if forecast > 1:
        crossings = _i.ta_rnn(crossings, range(1, forecast))

    return crossings.shift(-forecast)


def ta_future_bband_quantile(df: _pd.Series, period=5, forecast_period=5, stddev=2.0, ddof=1, include_mean=True):
    # we want to know if a future price is violating the current upper/lower band
    bands = _f.ta_bbands(df, period, stddev, ddof)
    bands = bands[["lower", "mean", "upper"] if include_mean else ["lower", "upper"] ]
    future = df.shift(-forecast_period)

    return bands \
        .join(future) \
        .apply(lambda row: _index_of_bucket(row[future.name], row[bands.columns]), axis=1, raw=False) \
        .rename(f"{df.name}_quantile")


def ta_future_multi_bband_quantile(df: _pd.Series, period=5, forecast_period=5, stddevs=[0.5, 1.0, 1.5, 2.0], ddof=1, include_mean=True):
    future = df.shift(-forecast_period)
    bands = _f.ta_multi_bbands(df, period, stddevs, ddof)

    if not include_mean:
        bands = bands.drop("mean", axis=1)

    return bands \
        .join(future) \
        .apply(lambda row: _index_of_bucket(row[future.name], row[bands.columns]), axis=1, raw=False) \
        .rename(f"{df.name}_quantile")


def ta_opening_gap(df: _pd.DataFrame, forecast_period=1, offset=0.005, open="Open", close="Close"):
    gap = (df[open].shift(-forecast_period) / df[close]) - 1
    return gap.apply(lambda row: _np.nan if _np.isnan(row) or _np.isinf(row) else \
                                 2 if row > offset else 1 if row < -offset else 0)\
              .rename("opening_gap")


def ta_opening_gap_closed(df: _pd.DataFrame, threshold=0.001, no_gap=_np.nan, open="Open", high="High", low="Low", close="Close"):
    gap = "$gap$"
    yesterday_close = "$close-1$"
    df = df[[open, high, low, close]].copy()
    df[yesterday_close] = df[close].shift(1)
    df[gap] = (df[open] / df[close].shift(1)) - 1

    # if open > close-1 we need to check if low <= close-1
    # if open < close-1 we need to check if high >= close-1

    def gap_category(row):
        if row[gap] < threshold:
            return no_gap
        else:
            if row[open] > row[yesterday_close]:
                return row[low] <= row[yesterday_close]
            elif row[open] < row[yesterday_close]:
                return row[high] >= row[yesterday_close]
            else:
                return no_gap

    return df.apply(gap_category, raw=False, axis=1).rename("closing_gap")



from typing import Union as _Union
from pyts.image import GramianAngularField as _GAF
import pandas as _pd
import numpy as _np
import logging
from pandas_ml_common import Typing, has_indexed_columns
from pandas_ml_common.utils import ReScaler
from pandas_ml_quant.analysis.encoders.auto_regression import ta_rnn

_log = logging.getLogger(__name__)


# possible encoders
def pyts_gaf(time_steps, **kwargs):
    image_size = max(min(kwargs["image_size"] if "image_size" in kwargs else time_steps, time_steps), 1)

    def encoder(x):
        gaf = _GAF(**kwargs)
        return gaf.fit_transform(x)

    return encoder


def invertible_gaf_encoder(time_steps, **kwargs):
    rescale = True if "rescale" in kwargs and kwargs["rescale"] else False

    I = _np.ones(time_steps)

    def encoder(x):
        if rescale:
            x = ReScaler((_np.min(x), _np.max(x)), (0, 1), clip=True)(x)

        a = _np.sqrt(I - x ** 2)
        gaf = _np.outer(x, x) - _np.outer(a, a)
        return gaf

    return lambda x: _np.apply_along_axis(encoder, 1, x)


# technical analysis function
def ta_gaf(df: Typing.PatchedPandas, columm_index_level=1, type='pyts', **kwargs):
    """
    :param df:
    :param columm_index_level:
    :param type:
    :param kwargs:
    :return: channel first 2D convolution GAF encoded matrix
    """

    if isinstance(df.columns, _pd.MultiIndex):
        # for each n'd level column create a dict like {Close: [(1, Close), (2, Close), ...]}
        l = columm_index_level
        columns = {c[l]: [c2 for c2 in df.columns.to_list() if c2[l] == c[l]] for c in df.columns.to_list()}
        res = _pd.DataFrame({}, index=df.index)

        # new we can loop feature for feature
        for feature, timesteps in columns.items():
            dff = df[timesteps]
            dff.columns = [t[:l] + t[l+1:] for t in timesteps]
            s = ta_gaf(dff, columm_index_level, type=type, **kwargs)
            res[f'{feature}_gaf'] = s

        return res
    else:
        time_steps = len(df.columns)

        if type == 'pyts':
            gaf_encoder = pyts_gaf(time_steps, **kwargs)
        elif type == 'invertible':
            if "rescale" not in kwargs or not kwargs["rescale"]:
                if _np.min(df.values) < 0 or _np.max(df.values) > 1:
                    _log.warning("data is not within [0, 1] and thus is not invertible!")

            gaf_encoder = invertible_gaf_encoder(time_steps, **kwargs)
        else:
            raise ValueError(f"unknown gaf encoder {type}! possible values ['pyts', 'invertible']")

        return _pd.Series(list(gaf_encoder(df.values)), name="gaf", index=df.index)


def ta_inverse_gasf(df: Typing.PatchedPandas):
    if has_indexed_columns(df):
        res = _pd.DataFrame({}, index=df.index)
        for col in df.columns:
            sdf = ta_inverse_gasf(df[col])
            sdf.columns = _pd.MultiIndex.from_product([[col], sdf.columns])
            res = res.join(sdf)

        return res

    inv = np_inverse_gaf(df._.values)
    if inv.ndim == 2:
        return _pd.DataFrame(inv, index=df.index)
    else:
        if inv.shape[1] == 1:
            return _pd.DataFrame(inv[:,0,:], index=df.index)
        else:
            raise ValueError("should never et here ..")


def np_inverse_gaf(values):
    # values have shape channel, w, h
    if values.ndim == 4:
        return _np.array([np_inverse_gaf(values[sample]) for sample in range(values.shape[0])])
    else:
        return _np.array([_inverse_gaf(values[channel]) for channel in range(values.shape[0])])


def _inverse_gaf(values):
    return _np.sqrt(((_np.diag(values) + 1) / 2))



import logging
import traceback

import numpy as np

from pandas_ml_quant.analysis import ta_ewma_covariance
from pandas_ml_common import pd, Typing
from qpsolvers import solve_qp
from statsmodels.stats.correlation_tools import cov_nearest



_log = logging.getLogger(__name__)


def ta_markowitz(df: Typing.PatchedDataFrame,
                 covariances=None,
                 risk_aversion=5,
                 return_period=60,
                 prices='Close',
                 expected_returns=None,
                 rebalance_trigger=None,
                 solver='cvxopt',
                 tail=None):
    assert isinstance(df.columns, pd.MultiIndex), \
        "expect multi index columns 'prices', 'expected returns' and rebalance trigger"

    # risk
    if covariances is None:
        cov = ta_ewma_covariance(df._[prices])
    elif isinstance(covariances, str):
        cov = df._[covariances]
    else:
        cov = covariances
    cov = cov.dropna()

    # return
    exp_ret = _default_returns_estimator(df, prices, expected_returns, return_period, len(cov.columns))
    exp_ret.columns = cov.columns

    # re-balance
    trigger = (pd.Series(np.ones(len(df)), index=df.index) if rebalance_trigger is None else df[rebalance_trigger]).dropna()

    # non negative weight constraint and weights sum to 1
    h = np.zeros(len(cov.columns)).reshape((-1, 1))
    G = -np.eye(len(h))
    A = np.ones(len(h)).reshape((1, -1))
    b = np.ones(1)

    # magic solution's
    keep_solution = (np.empty(len(h)) * np.nan)
    uninvest = np.zeros(len(h))

    # keep last solution
    last_solution = None

    def optimize(t, sigma, pi):
        nonlocal last_solution
        nr_of_assets = len(sigma)

        # only optimize if we have a re-balance trigger (early exit)
        if last_solution is not None and last_solution.sum() > 0.99:
            # so we had at least one valid solution in the past
            # we can early exit if we do not have any signal or or no signal for any currently hold asset
            if t.ndim > 1 and t.shape[1] == nr_of_assets:
                if t[:, last_solution >= 0.01].sum().any() < 1:
                    return keep_solution
            else:
                if t.sum().any() < 1:
                    return keep_solution

        # make sure covariance matrix is positive definite
        simga = cov_nearest(sigma)

        # we perform optimization except when all expected returns are < 0
        # then we early exit with an un-invest command
        if len(pi[:, pi[0] < 0]) == pi.shape[1]:
            return uninvest
        else:
            try:
                sol = solve_qp(risk_aversion * sigma, -pi.T, G=G, h=h, A=A, b=b, solver=solver)
                if sol is None:
                    _log.error("no solution found")
                    return uninvest
                else:
                    return sol
            except Exception as e:
                _log.error(traceback.format_exc())
                return uninvest

    index = sorted(set(df.index.intersection(cov.index.get_level_values(0)).intersection(exp_ret.index).intersection(trigger.index)))
    if tail is not None:
        index = index[-abs(tail)]

    weights = [optimize(trigger.loc[[i]].values, cov.loc[[i]].values, exp_ret[cov.columns].loc[[i]].values) for i in index]

    # turn weights into a data frame
    return pd.DataFrame(weights, index=index, columns=cov.columns)


def _default_returns_estimator(df, prices, expected_returns, return_period, nr_of_assets):
    # return
    if expected_returns is None:
        exp_ret = df._[prices].pct_change().rolling(return_period).mean()
    elif isinstance(expected_returns, (int, float, np.ndarray)):
        exp_ret = pd.Series(np.ones((len(df), nr_of_assets)) * expected_returns, index=df.index)
    elif isinstance(expected_returns, str):
        exp_ret = df._[expected_returns]
    else:
        exp_ret = expected_returns

    return exp_ret.dropna()



from typing import Callable

import numpy as np
import pandas as pd


PRICE = "price"
ASSET_VALUE = "asset_value"
POS_OPEN = "position_open"
POS_CLOSE = "position_close"
POS_NET = "position_net"
CASH_SLIPP = "slippage"
CASH_OPEN = "cash_open"
CASH_CLOSE = "cash_close"
CASH_NET = "cash_net"
NET = "net"


class TransactionLog(object):
    def __init__(self):
        super().__init__()
        self.transactions_open = {}
        self.transactions_close = {}

    def add_open_transaction(self, iloc: int, amount: float):
        amount = float(amount)

        if iloc in self.transactions_open:
            self.transactions_open[iloc] = [*self.transactions_open[iloc], amount]
        else:
            self.transactions_open[iloc] = [amount]

    def add_close_transaction(self, iloc: int, amount: float):
        amount = float(amount)

        if iloc in self.transactions_close:
            self.transactions_close[iloc] = [*self.transactions_close[iloc], amount]
        else:
            self.transactions_close[iloc] = [amount]

    def evaluate(self, prices: pd.Series, slippage: Callable[[float], float] = lambda value: 0):
        with pd.option_context('mode.chained_assignment', None):
            df = prices.to_frame()
            open_keys = [*self.transactions_open.keys()]
            close_keys = [*self.transactions_close.keys()]

            # calculate positions
            df[POS_OPEN] = 0
            df[POS_OPEN].iloc[open_keys] = [np.array([v]).sum() for v in self.transactions_open.values()]

            df[POS_CLOSE] = 0
            df[POS_CLOSE].iloc[close_keys] = [np.array([v]).sum() for v in self.transactions_close.values()]

            df[POS_NET] = (df[POS_OPEN] + df[POS_CLOSE]).cumsum()

            # calculate cash balance
            df[CASH_OPEN] = 0
            df[CASH_CLOSE] = 0
            df[CASH_SLIPP] = 0

            open = df[POS_OPEN].iloc[open_keys] * prices.iloc[open_keys]
            df[CASH_OPEN].iloc[open_keys] -= open

            close = df[POS_CLOSE].iloc[close_keys] * -prices.iloc[close_keys]
            df[CASH_CLOSE].iloc[close_keys] += close

            df[CASH_SLIPP].iloc[open_keys] += open.apply(slippage)
            df[CASH_SLIPP].iloc[close_keys] += close.apply(slippage)

            df[CASH_NET] = (df[CASH_OPEN] + df[CASH_CLOSE] + df[CASH_SLIPP]).cumsum()
            pd.set_option('display.max_columns', 500)

            # calculate position value
            df[ASSET_VALUE] = df[POS_NET] * prices
            df[NET] = df[CASH_NET] + df[ASSET_VALUE]

            return df


class StreamingTransactionLog(object):
    OPEN = 1
    CLOSE = -1

    def __init__(self):
        self.log = TransactionLog()
        self.current_index = 0
        self.current_position = None

    def rebalance(self, balance):
        if self.current_position is None:
            self.log.add_open_transaction(self.current_index, balance)
            self.current_position = 0
        else:
            if balance == 0:
                if self.current_position != 0:
                    self.log.add_close_transaction(self.current_index, -self.current_position)
            elif self.current_position > 0 and balance < 0:
                # swing long to short
                self.log.add_close_transaction(self.current_index, -self.current_position)
                self.log.add_open_transaction(self.current_index, balance)
            elif self.current_position < 0 and balance > 0:
                # swing short to long
                self.log.add_close_transaction(self.current_index, -self.current_position)
                self.log.add_open_transaction(self.current_index, balance)
            else:
                if balance > 0:
                    # long
                    delta = balance - self.current_position
                    if delta > 0:
                        self.log.add_open_transaction(self.current_index, delta)
                    else:
                        self.log.add_close_transaction(self.current_index, delta)
                else:
                    # short
                    delta = balance - self.current_position
                    if delta > 0:
                        self.log.add_close_transaction(self.current_index, delta)
                    else:
                        self.log.add_open_transaction(self.current_index, delta)

        self.current_index += 1
        self.current_position = balance

    def perform_action(self, action: int, amount):
        action = int(action)
        amount = float(amount)

        if action > 0:
            self.log.add_open_transaction(self.current_index, amount)
        elif action < 0:
            self.log.add_close_transaction(self.current_index, action)

        self.current_index += 1
        return self.current_index

    def evaluate(self, prices: pd.Series, slippage: Callable[[float], float] = lambda value: 0):
        return self.log.evaluate(prices, slippage)



from unittest import TestCase

import pandas_ml_quant
from pandas_ml_quant_test.config import DF_TEST

print(pandas_ml_quant.__version__)


class TestQuantComponent(TestCase):

    def test_all_analysis_functions(self):
        # df = DF_TEST.copy()
        #
        # ta_functions = [ta for ta in dir(df.ta) if ta.startswith("ta_")]
        # self.assertGreaterEqual(len(ta_functions), 44)
        #
        # # FIXME call all functions
        # df.ta.sma(20)
        pass



from unittest import TestCase

import numpy as np

from pandas_ml_quant.trading.transaction_log import TransactionLog, StreamingTransactionLog
from pandas_ml_quant_test.config import DF_TEST


class TestTransactionLog(TestCase):

    def test_balance_valuation_open_close(self):
        df = DF_TEST[-10:].copy()
        df["Price"] = range(1, 11)
        trans = TransactionLog()

        # long
        trans.add_open_transaction(2, 10)
        trans.add_close_transaction(4, -10)

        # short
        trans.add_open_transaction(7, -10)
        trans.add_close_transaction(8, 10)

        values = trans.evaluate(df["Price"], slippage=lambda _: -2)
        print(values)
        self.assertEqual(2, values.iloc[-1, -1])

    def test_buy_sell_same_day(self):
        df = DF_TEST[-10:].copy()
        df["Price"] = range(1, 11)
        trans = TransactionLog()

        # long
        trans.add_open_transaction(2, 10)
        trans.add_close_transaction(2, -10)

        trans.add_open_transaction(4, 10)
        trans.add_close_transaction(5, -5) # +10
        trans.add_close_transaction(6, -5) # + 5

        values = trans.evaluate(df["Price"])
        print(values)
        self.assertEqual(15, values.iloc[-1, -1])

    def test_open_close_same_day(self):
        df = DF_TEST[-10:].copy()
        df["Price"] = range(1, 11)
        trans = TransactionLog()

        # long
        trans.add_open_transaction(2, 10)
        trans.add_close_transaction(2, -2)
        trans.add_open_transaction(3, 2)
        trans.add_close_transaction(3, -1)
        trans.add_close_transaction(4, -9)

        values = trans.evaluate(df["Price"])
        print(values)

        self.assertListEqual([0, 0, -24, -24 -4,  17,  17,  17,  17,  17,  17], values["cash_net"].values.tolist())
        self.assertListEqual([0, 0, +24, +32 +4,   0,   0,   0,   0,   0,   0], values["asset_value"].values.tolist())
        self.assertListEqual([0, 0,   0,      8,  17,  17,  17,  17,  17,  17], values["net"].values.tolist())

    def test_balance_hf_open_close(self):
        df = DF_TEST[-10:].copy()
        df["Price"] = range(1, 11)
        trans = TransactionLog()

        # long 3 trades
        for i in range(2, 8, 2):
            trans.add_open_transaction(i, 10)
            trans.add_close_transaction(i + 1, -10)

        values = trans.evaluate(df["Price"])
        print(values)
        self.assertEqual(10 * 3, values.iloc[-1, -1])

    def test_balance_valuation_accumulate(self):
        df = DF_TEST[-10:].copy()
        df["Price"] = 10

        trans = TransactionLog()
        trans.add_open_transaction(2, 10)
        trans.add_open_transaction(4, 2)
        trans.add_close_transaction(5, -12)

        values = trans.evaluate(df["Price"])
        print(values)

    def test_rebalance(self):
        df = DF_TEST[-10:].copy()
        df["Price"] = range(1, 11)

        trans = StreamingTransactionLog()
        trans.rebalance(0.1)
        trans.rebalance(0)
        trans.rebalance(0.5)
        trans.rebalance(0.6)
        trans.rebalance(0.4)
        trans.rebalance(0.2)
        trans.rebalance(-0.2)
        trans.rebalance(-0.3)
        trans.rebalance(-0.1)
        trans.rebalance(0)

        values = trans.evaluate(df["Price"])
        print(values)
        self.assertEqual(0, values.iloc[-1, 1])
        np.testing.assert_almost_equal(1.2, values.iloc[-1, -1])



from pandas_ml_quant.trading.transaction_log import StreamingTransactionLog
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME, TARGET_COLUMN_NAME
from pandas_ml_utils.ml.summary import Summary
from pandas_ml_common import Typing


class BacktestSummary(Summary):

    def __init__(self, df: Typing.PatchedDataFrame, **kwargs):
        super().__init__(df, **kwargs)

    def calculate_performance(self, initial_capital=100, prediction_name="0", target_name="Close"):
        log = StreamingTransactionLog()
        self.df[PREDICTION_COLUMN_NAME, prediction_name].apply(lambda action: log.rebalance(action * initial_capital))
        perf = log.evaluate(self.df[TARGET_COLUMN_NAME, target_name].rename("prices"))

        return perf




from typing import Union

import pandas as pd
from sklearn.preprocessing import LabelBinarizer

from pandas_ml_common.utils import has_indexed_columns


def ta_one_hot_encode_discrete(po: Union[pd.Series, pd.DataFrame], drop_na=True) -> Union[pd.Series, pd.DataFrame]:
    if has_indexed_columns(po):
        return pd.DataFrame([ta_one_hot_encode_discrete(po[col]) for col in po.columns]).T
    else:
        if drop_na:
            po = po.dropna()

        values = po.values.astype(int)
        offset = po.min()
        values = values - offset
        nr_of_classes = values.max() + 1

        label_binarizer = LabelBinarizer()
        label_binarizer.fit(range(int(nr_of_classes)))
        return pd.Series(label_binarizer.transform(values).tolist(), index=po.index, name=po.name)





from unittest import TestCase
from pandas_ml_quant import np
from pandas_ml_quant.analysis.encoders import ta_gaf, ta_rnn, ta_inverse_gasf, np_inverse_gaf
from pandas_ml_quant_test.config import DF_TEST, DF_INVERSE_GAF
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME


class TestGAF(TestCase):

    def test_series(self):
        s = DF_TEST["Close"]
        timesteps = ta_rnn(s, [1, 2, 3])

        gaf = ta_gaf(timesteps)
        shape = gaf._.values.shape
        np.testing.assert_almost_equal(0.37, gaf.iloc[-1][0][0][0], 2)
        self.assertEqual((6760, 1, 3, 3), shape)

    def test_multi_channel(self):
        df = DF_TEST[["Open", "Close"]]
        timesteps = ta_rnn(df, [1, 2, 3])

        gaf = ta_gaf(timesteps)
        shape = gaf._.values.shape
        self.assertEqual((6760, 2, 3, 3), shape)

    def test_gasf_encode_decode(self):
        s = DF_TEST["Close"]
        timesteps = ta_rnn(s, [1, 2, 3]).ta.rescale((0, 1), axis=1)

        gasf = ta_gaf(timesteps, type='invertible', rescale=True)
        shape = gasf._.values.shape

        print(shape)
        print(gasf.tail())

        print(timesteps._.values[-2:])
        print(np_inverse_gaf(gasf._.values[-2:]))

        df = ta_inverse_gasf(gasf)
        print(df.tail())
        self.assertEqual((6760, 1, 3, 3), shape)
        # np.testing.assert_array_almost_equal(s, s_rec)

    def test_inverse_gaf(self):
        df = DF_INVERSE_GAF
        inv = ta_inverse_gasf(df[PREDICTION_COLUMN_NAME])

        print(inv)



from typing import Tuple, Callable

import pandas as pd

from pandas_ml_quant.trading.transaction_log import StreamingTransactionLog
from pandas_ml_common import Typing
from pandas_ml_common.utils import has_indexed_columns


def ta_backtest(signal: Typing.PatchedDataFrame,
                prices: Typing.PatchedPandas,
                action: Callable[[pd.Series], Tuple[int, float]],
                slippage: Callable[[float], float] = lambda _: 0):
    if has_indexed_columns(signal):
        assert len(signal.columns) == len(prices.columns), "Signal and Prices need the same shape!"
        res = pd.DataFrame({}, index=signal.index, columns=pd.MultiIndex.from_product([[], []]))

        for i in range(len(signal.columns)):
            df = ta_backtest(signal[signal.columns[i]], prices[prices.columns[i]], action, slippage)

            top_level_name = ",".join(prices.columns[i]) if isinstance(prices.columns[i], tuple) else prices.columns[i]
            df.columns = pd.MultiIndex.from_product([[top_level_name], df.columns.to_list()])
            res = res.join(df)

        return res

    assert isinstance(prices, pd.Series), "prices need to be a series!"
    trades = StreamingTransactionLog()

    def trade_log_action(row):
        direction_amount = action(row)
        if isinstance(direction_amount, tuple):
            trades.perform_action(*direction_amount)
        else:
            trades.rebalance(float(direction_amount))

    signal.to_frame().apply(trade_log_action, axis=1, raw=True)
    return trades.evaluate(prices.rename("price"), slippage)


# create convenient type hint
import numpy as _np

from pandas_ml_common import Typing


def ta_ewma_covariance(df: Typing.PatchedPandas, convert_to='returns', alpha=0.97):
    data = df.copy()

    if convert_to == 'returns':
        data = df.pct_change()
    if convert_to == 'log-returns':
        data = _np.log(df) - _np.log(df.shift(1))

    data.columns = data.columns.to_list()
    return data.ewm(com=alpha).cov()




from unittest import TestCase

from pandas_ml_quant.analysis import ta_decimal_year
from pandas_ml_quant_test.config import DF_TEST


class TestTimeEncoders(TestCase):

    def test_time_encoder(self):
        df = DF_TEST["Close"].copy()
        dec = ta_decimal_year(df)

        print(df.index[0], df.index[-1])
        print(dec[0], dec[-1])

        self.assertAlmostEqual(1993.0765027322404, dec[0])
        self.assertAlmostEqual(2019.9234972677596, dec[-1])




from typing import Union as _Union

# create convenient type hint
import numpy as _np
import pandas as _pd
from scipy.stats import zscore

from pandas_ml_common.utils import has_indexed_columns
from pandas_ml_quant.analysis.filters import ta_ema, ta_wilders, ta_sma
from pandas_ml_quant.utils import wilders_smoothing as _ws, with_column_suffix as _wcs

_PANDAS = _Union[_pd.DataFrame, _pd.Series]


def ta_inverse(df: _PANDAS) -> _PANDAS:
    return df.apply(lambda col: col * -1 + col.min() + col.max(), raw=True)




import logging
import os

import numpy as np
from matplotlib.figure import Figure

from pandas_ml_common import Typing
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.summary import ClassificationSummary

_log = logging.getLogger(__name__)


class BinaryWeightedClassificationSummary(ClassificationSummary):

    def __init__(self, df: Typing.PatchedDataFrame, probability_cutoff=0.5, **kwargs):
        super().__init__(df)
        pc = self.probability_cutoff = probability_cutoff

        self.true_neg = (df[PREDICTION_COLUMN_NAME].iloc[:, 0] <  pc) & (df[LABEL_COLUMN_NAME].iloc[:, 0] < pc)
        self.true_pos = (df[PREDICTION_COLUMN_NAME].iloc[:, 0] >  pc) & (df[LABEL_COLUMN_NAME].iloc[:, 0] > pc)
        self.pos_idx = self.true_pos | self.true_neg

        self.false_pos = (df[PREDICTION_COLUMN_NAME].iloc[:, 0] >= pc) & (df[LABEL_COLUMN_NAME].iloc[:, 0] < pc)
        self.false_neg = (df[PREDICTION_COLUMN_NAME].iloc[:, 0] <  pc) & (df[LABEL_COLUMN_NAME].iloc[:, 0] > pc)
        self.neg_idx = self.false_pos | self.false_neg

    def set_probability_cutoff(self, probability_cutoff: float):
        return BinaryWeightedClassificationSummary(self.df, probability_cutoff, **self.kwargs)

    def gross_confusion(self):
        return (
            len(self.df),
            self.df[GROSS_LOSS_COLUMN_NAME][self.pos_idx].count().values.sum(),
            self.df[GROSS_LOSS_COLUMN_NAME][self.neg_idx].dropna().abs().values.sum()
        )

    def plot_classification(self, figsize=(16, 9)) -> Figure:
        import seaborn as sns
        import matplotlib.pyplot as plt
        from pandas.plotting import register_matplotlib_converters

        # get rid of deprecation warning
        register_matplotlib_converters()
        pc = self.probability_cutoff
        df = self.df

        # define grid
        fig, ax1 = plt.subplots(1, 1, figsize=figsize)

        x = np.arange(len(df))
        yp = df[PREDICTION_COLUMN_NAME].iloc[:, 0]
        yl = df[GROSS_LOSS_COLUMN_NAME].iloc[:, 0]

        # plot true positives/negatives
        sns.scatterplot(ax=ax1,
                        x=x[self.pos_idx],
                        y=yl[self.pos_idx],
                        size=np.abs(yp[self.pos_idx].values - 0.5),
                        size_norm=(0, 0.5),
                        color=sns.xkcd_rgb['pale green'],
        )

        # plot false positives/negatives
        sns.scatterplot(ax=ax1,
                        x=x[self.neg_idx],
                        y=yl[self.neg_idx],
                        size=np.abs(yp[self.neg_idx].values - 0.5),
                        size_norm=(0, 0.5),
                        color=sns.xkcd_rgb['cerise'],
        )

        # make y symmetric
        maax = np.abs(yl.dropna().values).max()
        ax1.set_ylim((-maax, maax))

        # add mean line
        ax1.hlines(0, 0, len(df), color=sns.xkcd_rgb['silver'])

        return fig

    def _repr_html_(self):
        from mako.template import Template

        path = os.path.dirname(os.path.abspath(__file__))
        file = os.path.join(path, 'html', 'binary_weighted_classification_summary.html')
        template = Template(filename=file)

        return template.render(
            summary=self,
            gross_confusion=self.gross_confusion(),
            roc_plot=plot_to_html_img(self.plot_ROC),
            cmx_plot=plot_to_html_img(self.plot_confusion_matrix),
            gross_loss_plot=plot_to_html_img(self.plot_classification),
        )


class MultipleBinaryWeightedClassificationSummary(ClassificationSummary):

    def __init__(self, df: Typing.PatchedDataFrame, probability_cutoff=0.5, **kwargs):
        super().__init__(df)
        df = df[[LABEL_COLUMN_NAME, PREDICTION_COLUMN_NAME, GROSS_LOSS_COLUMN_NAME]]
        nr_columns = len(df.columns) // 3

        self.binary_summaries =\
            [BinaryWeightedClassificationSummary(df.iloc[:,[i, i+nr_columns, i+nr_columns*2]], probability_cutoff, **kwargs)
             for i in range(nr_columns)]

    def set_probability_cutoff(self, probability_cutoff: float):
        return MultipleBinaryWeightedClassificationSummary(self.df, probability_cutoff, **self.kwargs)

    def _repr_html_(self):
        from mako.template import Template

        path = os.path.dirname(os.path.abspath(__file__))
        file = os.path.join(path, 'html', 'multi_binary_weighted_classification_summary.html')
        template = Template(filename=file)

        return template.render(
            summary=self,
            gross_confusions=[bs.gross_confusion() for bs in self.binary_summaries],
            roc_plots=[plot_to_html_img(bs.plot_ROC) for bs in self.binary_summaries],
            cmx_plots=[plot_to_html_img(bs.plot_confusion_matrix) for bs in self.binary_summaries],
            # gross_loss_plot=[plot_to_html_img(bs.plot_classification) for bs in self.binary_summaries],
        )




"""Augment pandas DataFrame with methods for quant analysis"""
__version__ = '0.1.8'

import importlib
import sys as _sys

from pandas_ml_common import *
from pandas_ml_quant.df.technical_analysis import TechnicalAnalysis as _TA
from pandas_ml_quant.model import *

_log = logging.getLogger(__name__)
_log.debug(f"numpy version {np.__version__}")
_log.debug(f"pandas version {pd.__version__}")


setattr(PandasObject, "ta", property(lambda self: _TA(self)))

# auto import pandas ml utils
if 'pandas_ml_utils' in _sys.modules:
    pandas_ml_utils = importlib.import_module("pandas_ml_utils")
    _log.warning(f"automatically imported pandas_ml_utils {pandas_ml_utils.__version__}")

# auto import pandas ml quant data prviders
if 'pandas_ml_quant_data_provider' in _sys.modules:
    pandas_ml_quant_data_provider = importlib.import_module("pandas_ml_quant_data_provider")
    _log.warning(f"automatically imported pandas_ml_quant_data_provider {pandas_ml_quant_data_provider.__version__}")
else:
    _log.warning("pandas_ml_quant_data_provider module not avialable!")




import logging
from typing import Tuple

from gym.spaces import MultiDiscrete, Box, Discrete

from pandas_ml_common import np, pd
from pandas_ml_quant.trading.transaction_log import StreamingTransactionLog
from pandas_ml_utils import ReinforcementModel

_log = logging.getLogger(__name__)


class TradingAgentGym(ReinforcementModel.DataFrameGym):

    def __init__(self,
                 input_shape: Tuple[int, ...],
                 trading_fraction: int = 10,
                 trading_assets: int = 1,  # later we want the bot to trade one of multiple possible assets
                 allow_short: bool = False,
                 stop_if_lost: float = None,
                 initial_capital: float = 100000,
                 commission=lambda size: 0.025):
        super().__init__(MultiDiscrete([trading_assets, trading_fraction]) if trading_assets > 1 else
                         Discrete(trading_fraction + 1),
                         Box(low=-1, high=1, shape=input_shape)) # FIXME what shape? we also need historic trades?

        self.trading_fraction = trading_fraction
        self.initial_capital = initial_capital
        self.commission = commission
        self.stop_if_lost = stop_if_lost
        self.allow_short = allow_short

        if allow_short and (trading_fraction % 2) != 0:
            _log.warning('short trades expect even nr of trading fraction')

        # eventually do not serialize ..
        self.trade_log = StreamingTransactionLog()
        self.current_net = 0

    def reset(self) -> np.ndarray:
        self.trade_log = StreamingTransactionLog()
        self.current_net = 0
        return super().reset()

    def interpret_action(self,
                         action,
                         idx: int,
                         features: np.ndarray,
                         labels: np.ndarray,
                         targets: np.ndarray,
                         weights: np.ndarray = None,
                         gross_loss: np.ndarray = None) -> float:
        if self.allow_short:
            # 0 - 10  ->  -10 - 10
            action -= int(self.trading_fraction / 2) * 2

        # we convert the action in to a target balance we pass to the transaction log
        balance = action / self.trading_fraction / float(targets) * self.initial_capital
        return balance

    def take_action(self,
                    action,
                    idx: int,
                    features: np.ndarray,
                    labels: np.ndarray,
                    targets: np.ndarray,
                    weights: np.ndarray = None,
                    gross_loss: np.ndarray = None) -> float:

        # skip the very first bar
        if idx <= 0:
            self.trade_log.rebalance(0)
            return 0

        # we use a n/fractions as target balance -> 10 shares 20,... shares, ...
        balance = action / self.trading_fraction * self.initial_capital
        self.trade_log.rebalance(balance)

        # new we need to evaluate the portfolio performance
        # where the targets are the prices
        prices = pd.Series(self.train[2][:idx+1].ravel(), name="prices")
        perf = self.trade_log.evaluate(prices, self.commission)
        self.current_net = perf["net"].iloc[-1]

        if self.stop_if_lost is not None and self.current_net < self.initial_capital * (1 - self.stop_if_lost):
            raise StopIteration(f"lost more then {self.stop_if_lost}%")

        return self.calculate_trade_reward(perf)

    def calculate_trade_reward(self, portfolio_performance_log) -> float:
        return portfolio_performance_log["net"].iloc[-1]

    def next_observation(self,
                         idx: int,
                         features: np.ndarray,
                         labels: np.ndarray,
                         targets: np.ndarray,
                         weights: np.ndarray = None,
                         gross_loss: np.ndarray = None) -> np.ndarray:
        # currently returns only the features, but we also want to return some net worth, history, ...
        # for CNNs input must be [0, 255] and channel last, this means we need to reshape our feature space
        # return features.swapaxes(0, 1).swapaxes(1, 2) * 255
        pass

    def render(self, mode='human'):
        if mode == 'system':
            print(self.current_net)
        elif mode == 'notebook':
            # TODO plot something using matplotlib
            pass
        elif mode == 'human':
            # TODO plot something using matplotlib
            pass




import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import ipywidgets as wg
import numpy as np
import pandas as pd
from matplotlib import gridspec
from moviepy.video.io.bindings import mplfig_to_npimage
from pandas.plotting import register_matplotlib_converters
from datetime import timedelta
from pandas_ml_common.plot import plot_bar, plot_stacked_bar, plot_candlestick, plot_line, plot_matrix
from pandas_ml_common.plot.animations import plot_animation
from pandas_ml_common import get_pandas_object
from pandas_ml_quant.analysis import ta_trend_lines

register_matplotlib_converters()


class TaPlot(object):

    def __init__(self, df: pd.DataFrame, figsize, rows=2, cols=1, main_height_ratio=4):
        fig = plt.figure('r-', figsize=figsize)
        grid = gridspec.GridSpec(rows, cols, height_ratios=[main_height_ratio, *[1 for _ in range(1, rows)]])
        axis = []

        for i, gs in enumerate(grid):
            ax = fig.add_subplot(gs, sharex=axis[0] if i > 0 else None)
            ax.xaxis_date()
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))

            if i < rows - 1:
                ax.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
            else:
                ax.tick_params(axis='x', labelrotation=45)

            axis.append(ax)

        plt.xticks(rotation=45)

        self.df = df
        self.x = mdates.date2num(df.index)
        self.fig = fig
        self.axis = axis
        self.grid = grid

    def candlestick(self, open="Open", high="High", low="Low", close="Close", panel=0):
        self.axis[panel] = plot_candlestick(self.df, open, high, low, close, ax=self.axis[panel])
        return self._return()

    def stacked_bar(self, columns, padding=0.02, panel=1, **kwargs):
        # todo if x axis is multilevel then stack all bars at level > 2
        self.axis[panel] = plot_stacked_bar(self.df, columns, ax=self.axis[panel], padding=padding, **kwargs)
        return self._return()

    def bars(self):
        # FIXME add side by side bars
        # todo if x axis is multilevel then stack all bars at level > 2
        pass

    def bar(self, fields="Volume", panel=1, colors=None, color_map: str = 'afmhot', **kwargs):
        self.axis[panel] = plot_bar(self.df, fields, ax=self.axis[panel], colors=colors, color_map=color_map, **kwargs)
        return self._return()

    def line(self, fields="Close", panel=0, oscillator=False, **kwargs):
        self.axis[panel] = plot_line(self.df, fields, ax=self.axis[panel], **kwargs)
        return self._return()

    def plot_matrix(self, panel, fields, **kwargs):
        plot_matrix()

    def plot_matrix_animation(self, fields, fps=2, **kwargs):
        def make_frame(index):
            fig, ax = plt.subplots(figsize=(9, 9))
            plot_matrix(self.df.loc[[index]], fields, ax=ax, **kwargs)
            frame = mplfig_to_npimage(fig)
            plt.close(fig)
            return frame

        plt.close(self.fig)
        return plot_animation(self.df.index, make_frame, fps, **kwargs)

    def set_scale(self, panel, min, max):
        self.axis[panel].set_ylim(min, max)
        return self._return()

    def with_defaults(self):
        self.line()
        self.bar()
        return self.with_legend()

    def with_trend_lines(self,
                         field="Close",
                         panel=0,
                         edge_periods=3,
                         rescale_digits=4,
                         degrees=(-90, 90),
                         angles=30,
                         rho_digits=2):
        plt.close(self.fig)
        accumulation, lookup =\
            ta_trend_lines(get_pandas_object(self.df, field), edge_periods, rescale_digits, degrees, angles, rho_digits)

        def plot_trend_line(time, touches):
            ax = self.axis[panel]
            td = timedelta(days=time[0]), timedelta(days=time[1])

            # first remove all previous trend lines
            ax.lines = [l for l in ax.lines if not l.get_label().startswith(".Trend")]

            # then select the lines from the lookup table
            filtered = lookup[(lookup["touch"] >= touches[0]) & (lookup["touch"] <= touches[1])]
            filtered = filtered[(filtered["distance"] >= td[0]) & (filtered["distance"] <= td[1])]

            for i, tl in filtered.iterrows():
                points = tl["points"][0], tl["points"][-1]
                ax.plot([p[0] for p in points], [p[1] for p in points], label=".Trend")

            return self.fig

        #  TODO later add a wg.IntSlider to extend the trend lines from ots last point
        min_ts, max_ts = 2, len(self.df)
        time_silder = wg.IntRangeSlider(value=[max_ts, max_ts], min=min_ts, max=max_ts, step=1,
                                        continuous_update=False, description='Period:')

        min_to, max_to = 2, lookup["touch"].max()
        touch_silder = wg.IntRangeSlider(value=[min_to, max_to], min=min_to, max=max_to, step=1,
                                         continuous_update=False, description='Touches:')

        wg.interact(plot_trend_line, time=time_silder, touches=touch_silder)
        self.fig.show()
        return self

    def with_symetric_scale(self, *panels):
        for panel in panels:
            yl = np.abs(np.array(list(self.axis[panel].get_ylim()))).max() * 1.1
            self.axis[panel].set_ylim(-yl, +yl)

        return self._return()

    def __call__(self, *args, **kwargs):
        """
        TODO i am thinking of something like df.ta.plot()(candlesticks=True, sma=200)
             ... more thinking needed

        :param args:
        :param kwargs:
        :return:
        """
        if "lines" in kwargs:
            self.line(kwargs.pop('lines', None), **kwargs)
        else:
            self.line()

        if "bars" in kwargs:
            self.bar(kwargs.pop('bars', None), **kwargs)
        else:
            self.bar()

        return self.with_legend()

    def with_legend(self):
        for panel in self.axis:
            panel.legend(loc=2)

        return self

    def _return(self):
        self.grid.tight_layout(self.fig)


# %matplotlib
#inline
#
#from moviepy.editor import VideoClip
#from moviepy.video.VideoClip import DataVideoClip
#from moviepy.video.io.bindings import mplfig_to_npimage
#
#samples = gaf._.values[:, -1][-20:]
#
#
#def make_frame(sample):
#    fig, ax_patchwork = plt.subplots(figsize=(9, 9))
#
#    ax_patchwork.matshow(sample)
#    ax_patchwork.set_title("Gramian Angular Field")
#    ax_patchwork.set_yticklabels([])
#    ax_patchwork.set_xticklabels([])
#
#    frame = mplfig_to_npimage(fig)
#    plt.close(fig)
#
#    return frame
#
#
# # animation = VideoClip(make_frame, duration=5.1)
#animation = DataVideoClip(samples, make_frame, fps=4)
#
# # if(write_gif):
# # animation.write_gif("/tmp/foo.gif",fps=1)
#
#animation.ipython_display()
#


import logging
from unittest import TestCase

from pandas_ml_quant.analysis.indicators import ta_zscore
from pandas_ml_quant.trading.strategy.optimized import ta_markowitz
from pandas_ml_quant_test.config import DF_TEST_MULTI
from pandas_ml_common import pd, np, inner_join

logging.basicConfig(level=logging.DEBUG)


class TestOptimizedStrategies(TestCase):

    def test__ewma_markowitz(self):
        """given"""
        df = DF_TEST_MULTI

        """when"""
        portfolios = ta_markowitz(df, return_period=20)

        """then"""
        print(portfolios)
        np.testing.assert_array_almost_equal(np.array([0.683908, 3.160920e-01]), portfolios.iloc[-4].values, 0.00001)

    def test__given_expected_returns(self):
        # _default_returns_estimator
        pass

    def test__markowitz_strategy(self):
        """given"""
        df = DF_TEST_MULTI
        df_price = df._['Close']
        df_expected_returns = df_price._["Close"].ta.macd()._["histogram"]
        df_trigger = ta_zscore(df_price['Close']).abs() > 2.0
        df_data = inner_join(df_price, df_expected_returns)
        df_data = inner_join(df_data, df_trigger, prefix='trigger')

        """when"""
        portfolios = ta_markowitz(df_data,
                                  prices='Close',
                                  expected_returns='histogram',
                                  rebalance_trigger='trigger')

        """then"""
        print(portfolios)
        np.testing.assert_array_almost_equal(np.array([1.000000, 3.904113e-07]), portfolios.iloc[-1].values, 0.00001)



import pandas as _pd

# create convenient type hint
from pandas_ml_common import Typing as _t


def ta_volume_as_time(df: _t.PatchedPandas, volume="Volume"):
    if df.has_indexed_columns():
        res = df.copy()
        res.index = df[volume].cumsum()
        return res
    else:
        return _pd.Series(df.index, index=df.cumsum())


def ta_volume_interpolated():
    # TODO ...
    # first need a line from one (volume, price) to the next (volume, price)
    # then we can do so as if we would have evenly spaced volumes and calculate the interpolated price
    # finally we can map back the interpolated price to time
    pass



from keras.layers import Layer, Concatenate
import tensorflow as tf


class Time2Vec(Layer):
    """
    source: https://towardsdatascience.com/time2vec-for-time-series-features-encoding-a03a4f3f937e
    and:    https://arxiv.org/pdf/1907.05321.pdf
    """

    def __init__(self, output_dim=None, **kwargs):
        self.output_dim = output_dim
        super().__init__(**kwargs)

    def build(self, input_shape):

        self.W = self.add_weight(name='W',
                                 shape=(self.output_dim,
                                        self.output_dim),
                                 initializer='uniform',
                                 trainable=True)

        self.B = self.add_weight(name='B',
                                 shape=(input_shape[1],
                                        self.output_dim),
                                 initializer='uniform',
                                 trainable=True)

        self.w = self.add_weight(name='w',
                                 shape=(1, 1),
                                 initializer='uniform',
                                 trainable=True)

        self.b = self.add_weight(name='b',
                                 shape=(input_shape[1], 1),
                                 initializer='uniform',
                                 trainable=True)

        super().build(input_shape)

    def call(self, x, **kwargs):
        from keras import backend as K

        original = self.w * x + self.b
        x = K.repeat_elements(x, self.output_dim, -1)
        sin_trans = K.sin(K.dot(x, self.W) + self.B)
        return K.concatenate([sin_trans ,original], -1)

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[1], self.output_dim +1


class tf_Time2Vec(tf.keras.layers.Layer):
    """
    source: https://towardsdatascience.com/time2vec-for-time-series-features-encoding-a03a4f3f937e
    and:    https://arxiv.org/pdf/1907.05321.pdf
    """

    def __init__(self, output_dim=None, **kwargs):
        self.output_dim = output_dim
        super().__init__(**kwargs)

    def build(self, input_shape):

        self.W = self.add_weight(name='W',
                                 shape=(self.output_dim,
                                        self.output_dim),
                                 initializer='uniform',
                                 trainable=True)

        self.B = self.add_weight(name='B',
                                 shape=(input_shape[1].value,
                                        self.output_dim),
                                 initializer='uniform',
                                 trainable=True)

        self.w = self.add_weight(name='w',
                                 shape=(1, 1),
                                 initializer='uniform',
                                 trainable=True)

        self.b = self.add_weight(name='b',
                                 shape=(input_shape[1].value, 1),
                                 initializer='uniform',
                                 trainable=True)

        super().build(input_shape)

    def call(self, x, **kwargs):
        K = tf.keras.backend

        original = self.w * x + self.b
        x = K.repeat_elements(x, self.output_dim, -1)
        sin_trans = K.sin(K.dot(x, self.W) + self.B)
        return K.concatenate([sin_trans, original], -1)

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[1], self.output_dim +1



from unittest import TestCase

import numpy as np
import torch as t
import torch.nn as nn

from pandas_ml_quant.pytorch.layers import Time2Vec
from pandas_ml_quant_test.config import DF_TEST


class TestTime2Vec(TestCase):

    def test_time2vec(self):
        df = DF_TEST.copy()

        x = df["Volume"].pct_change().ta.rnn(range(5))._.values[:-2]
        y = df["Volume"].pct_change().shift(-1).dropna()[6:]

        class MyModule(nn.Module):

            def __init__(self, input, output):
                super().__init__()
                self.net = Time2Vec(input, output)

            def forward(self, x):
                return self.net(x)

        t2v = MyModule(5, 32)
        print(x.shape)
        res = t2v(t.from_numpy(x).float())
        print(res)
        print(res.shape)  # -> ?, 5, 33

    def _test_lala(self):
        class MyModule(nn.Module):

            def __init__(self, input, output):
                super().__init__()
                self.net = Time2Vec(input, output)

            def forward(self, x):
                return self.net(x)

        t2v = MyModule(3, 3)
        print(t2v(t.from_numpy(np.array([[[0.1], [0.2], [0.3]], [[0.1], [0.2], [0.3]]])).float()))


import numpy as np
import pandas as pd
from scipy.stats import norm

from pandas_ml_common import Typing, has_indexed_columns
from pandas_ml_common.utils import ReScaler
from pandas_ml_quant.analysis import filters as _f
from pandas_ml_quant.utils import with_column_suffix as _wcs


def ta_rescale(df: pd.DataFrame, range=(-1, 1), digits=None, axis=None):
    if axis is not None:
        return df.apply(lambda x: ta_rescale(x, range), raw=False, axis=axis, result_type='broadcast')
    else:
        rescaler = ReScaler((df.values.min(), df.values.max()), range)
        rescaled = rescaler(df.values)

        if digits is not None:
            rescaled = np.around(rescaled, digits)

        if rescaled.ndim > 1:
            return pd.DataFrame(rescaled, columns=df.columns, index=df.index)
        else:
            return pd.Series(rescaled, name=df.name, index=df.index)


def ta_returns(df: Typing.PatchedPandas):
    return _wcs("return", df.pct_change())


def ta_log_returns(df: Typing.PatchedPandas):
    current = df
    lagged = df.shift(1)

    return _wcs("log_return", np.log(current) - np.log(lagged))


def ta_ma_ratio(df: Typing.PatchedPandas, period=20, lag=0, ma='sma', **kwargs):
    mafunc = getattr(_f, f'ta_{ma}')
    return _wcs(f"{ma}({period}) x 1/", df / mafunc(df, period=period, **kwargs).shift(lag).values - 1, df)


def ta_ncdf_compress(df: Typing.PatchedPandas, period=200, lower_percentile=25, upper_percentile=75) -> Typing.PatchedPandas:
    if has_indexed_columns(df):
        return pd.DataFrame(
            {col: ta_ncdf_compress(df[col], period, lower_percentile, upper_percentile) for col in df.columns},
            index=df.index
        )

    f50 = df.rolling(period).mean().rename("f50")
    fup = df.rolling(period).apply(lambda x: np.percentile(x, upper_percentile)).rename("fup")
    flo = df.rolling(period).apply(lambda x: np.percentile(x, lower_percentile)).rename("flo")

    return pd.Series(norm.cdf(0.5 * (df - f50) / (fup - flo)) - 0.5, index=df.index, name=df.name)


def ta_z_norm(df: Typing.PatchedPandas, period=200, ddof=1, demean=True, lag=0):
    if has_indexed_columns(df):
        return pd.DataFrame(
            {col: ta_z_norm(df[col], period, ddof, demean) for col in df.columns},
            index=df.index
        )

    # (value - mean) / std
    s = df.rolling(period).std(ddof=ddof)
    a = (df - df.rolling(period).mean().shift(lag)) if demean else df
    return (a / s / 4).rename(df.name)


def ta_performance(df: Typing.PatchedPandas):
    delta = df.pct_change() + 1
    return delta.cumprod()


def _ta_adaptive_normalisation():
    # TODO implement .papers/Adaptive Normalization.pdf
    pass